"""Testing FeatureAnnotation component
"""

import os
import sys
import threading
sys.path.append('./src/')
from fusion_tools.visualization import Visualization
from fusion_tools.handler.dsa_handler import DSAHandler
from fusion_tools.components import SlideMap, FeatureAnnotation, BulkLabels, SlideAnnotation, SlideAnnotationSchema

import pandas as pd

test_schema = SlideAnnotationSchema(
    {
        "name": "Test Slide Annotation",
        "description": "This is a test of the SlideAnnotation component",
        "annotations": [
            {
                "name": "Example Text Label",
                "description": "This is an example free text annotation type.",
                "type": "text",
                "roi": True,
                "editable": True
            },
            {
                "name": "Example Numeric Label",
                "description": "This is an example numeric input label",
                "type": "numeric",
                "min": 0,
                "max": 100,
                "roi": False
            },
            {
                "name": "Example Options Label",
                "description": "This is an example of a label with predefined options",
                "type": "options",
                "options": [
                    "Option 1",
                    "Option 2",
                    "Option 3"
                ],
                "multi": False,
                "roi": False
            }
        ]
    }
)

test_feature_schema = {
    'classes': [
        {
            'name': 'Test Class',
            'color': 'rgb(255,0,0)'
        }
    ],
    'labels': [
        {
            'name': 'Test Text Label',
            'type': 'text'
        },
        {
            'name': 'Test Options Label',
            'type': 'options',
            'options': [
                'Option 1',
                'Option 2'
            ]
        }
    ]
}

def main():

    # Grabbing first item from demo DSA instance
    #base_url = 'https://demo.kitware.com/histomicstk/api/v1'
    #item_id = ['5bbdeed1e629140048d01bcb','58b480ba92ca9a000b08c89d']
    base_url = os.environ.get('DSA_URL')

    item_id = [
        '6495a4e03e6ae3107da10dc5',
        '6495a4df3e6ae3107da10dc2'
    ] 


    # Starting the DSAHandler to grab information:
    dsa_handler = DSAHandler(
        girderApiUrl = base_url
    )
    
    vis_session = Visualization(
        tileservers=[dsa_handler.get_tile_server(i) for i in item_id],
        components = [
            [
                SlideMap(),
                [
                    FeatureAnnotation(
                        storage_path = os.getcwd()+'\\tests\\Test_Annotations\\',
                        labels_format = 'json',
                        annotations_format = 'rgb',
                        preset_schema = test_feature_schema
                    ),
                    BulkLabels(),
                    SlideAnnotation(
                        preload_schema=test_schema
                    )
                ]
            ]
        ]
    )

    vis_session.start()


if __name__=='__main__':
    main()





"""Testing CustomFunction component
"""

import os
import sys
import threading
sys.path.append('./src/')
from fusion_tools.visualization import Visualization
from fusion_tools.handler.dsa_handler import DSAHandler
from fusion_tools.components import SlideMap, CustomFunction, FUSIONFunction, OverlayOptions

import geojson
import rasterio.features
import geopandas as gpd

from skimage.filters import threshold_otsu

import plotly.express as px
import plotly.graph_objects as go
from matplotlib import cm
from PIL import Image
from skimage.transform import resize

from shapely.geometry import shape
import numpy as np
from skimage.segmentation import watershed
from skimage.color import rgb2hsv
import scipy.ndimage as ndi
from skimage.morphology import remove_small_holes, remove_small_objects
from skimage import exposure
from skimage.feature import peak_local_max

from io import BytesIO
import base64

# function that just expands geometry boundaries by a certain amount
def buffer_shapes(feature, buffer_radius):

    feature_shape = shape(feature['geometry'])
    buffered_shape = feature_shape.buffer(buffer_radius)

    return buffered_shape.__geo_interface__

# Defining CustomFunction component for buffering shapes
buffer_shapes_component = FUSIONFunction(
    title = 'Buffer Shapes',
    description = 'For a given annotation, perform a buffer operation by a specified value.',
    urls = [
        'https://shapely.readthedocs.io/en/stable/manual.html#constructive-methods'
    ],
    function = lambda feature,radius: buffer_shapes(feature,radius),
    function_type = 'forEach',
    input_spec = [
        {
            'name': 'feature',
            'type': 'annotation',
            'property': 'feature'
        },
        {
            'name': 'buffer_radius',
            'type': 'numeric',
            'min': -10,
            'max': 100
        }
    ],
    output_spec = [
        {
            'name': 'Buffered Annotation',
            'description': 'Buffered Annotation Boundaries',
            'type': 'annotation'
        }
    ],
)



# define sub-compartment segmentation function
def stain_mask(image, mask, nuc_thresh, eosin_thresh):

    seg_params = [
        {
            'name': 'Nuclei',
            'threshold': nuc_thresh,
            'min_size': 5
        },
        {
            'name': 'Eosinophilic',
            'threshold': eosin_thresh,
            'min_size': 10
        },
        {
            'name': 'Luminal Space',
            'threshold': 0,
            'min_size': 0
        }
    ]

    image_shape = np.shape(image)
    sub_comp_image = np.zeros((image_shape[0],image_shape[1],3))
    remainder_mask = np.ones((image_shape[0],image_shape[1]))

    hsv_image = np.uint8(255*rgb2hsv(image))
    hsv_image = hsv_image[:,:,1]
    #hsv_image = np.uint8(255*exposure.equalize_adapthist(hsv_image))
    #hsv_image = np.uint8(255*gaussian(hsv_image,sigma=0.5))

    p2, p98 = np.percentile(hsv_image, (2, 98))
    hsv_image = exposure.rescale_intensity(hsv_image, in_range=(p2, p98))

    for idx,param in enumerate(seg_params):

        # Check for if the current sub-compartment is nuclei
        if param['name'].lower()=='nuclei':

            remaining_pixels = np.multiply(hsv_image,remainder_mask)
            masked_remaining_pixels = np.multiply(remaining_pixels,mask)

            # Applying manual threshold
            masked_remaining_pixels[masked_remaining_pixels<=param['threshold']] = 0
            masked_remaining_pixels[masked_remaining_pixels>0] = 1

            # Area threshold for holes is controllable for this
            sub_mask = remove_small_holes(masked_remaining_pixels>0,area_threshold=10)
            sub_mask = sub_mask>0

            # Watershed implementation from: https://scikit-image.org/docs/stable/auto_examples/segmentation/plot_watershed.html
            distance = ndi.distance_transform_edt(sub_mask)
            labeled_mask, _ = ndi.label(sub_mask)
            coords = peak_local_max(distance,footprint=np.ones((3,3)),labels = labeled_mask)
            watershed_mask = np.zeros(distance.shape,dtype=bool)
            watershed_mask[tuple(coords.T)] = True
            markers, _ = ndi.label(watershed_mask)
            sub_mask = watershed(-distance,markers,mask=sub_mask)
            sub_mask = sub_mask>0

            # Filtering out small objects again
            sub_mask = remove_small_objects(sub_mask,param['min_size'])

        else:

            remaining_pixels = np.multiply(hsv_image,remainder_mask)
            masked_remaining_pixels = np.multiply(remaining_pixels,mask)

            # Applying manual threshold
            masked_remaining_pixels[masked_remaining_pixels<=param['threshold']] = 0
            masked_remaining_pixels[masked_remaining_pixels>0] = 1

            # Filtering by minimum size
            small_object_filtered = (1/255)*np.uint8(remove_small_objects(masked_remaining_pixels>0,param['min_size']))

            sub_mask = small_object_filtered

        sub_comp_image[sub_mask>0,idx] = 1
        remainder_mask -= sub_mask>0

    # Assigning remaining pixels within the boundary mask to the last sub-compartment
    remaining_pixels = np.multiply(mask,remainder_mask)
    #remaining_pixels = remainder_mask
    sub_comp_image[remaining_pixels>0,idx] = 1

    final_mask = np.zeros_like(remainder_mask)
    final_mask += sub_comp_image[:,:,0]
    final_mask += 2*sub_comp_image[:,:,1]
    final_mask += 3*sub_comp_image[:,:,2]

    final_mask[~mask] = 0
    return final_mask

# Defining CustomFunction component for sub-compartment segmentation
stain_mask_component = FUSIONFunction(
    title = 'Sub-Compartment Segmentation',
    description = 'For a given structure, use image analysis to segment different sub-compartments',
    urls = [
        'https://github.com/SarderLab/HistoLens'
    ],
    function = lambda image,mask,nuc_thresh,eosin_thresh: stain_mask(image,mask,nuc_thresh,eosin_thresh),
    function_type = 'forEach',
    input_spec = [
        {
            'name': 'image',
            'description': 'Raw RGB image containing IHC-stained histology',
            'type': 'image',
            'property': 'image'
        },
        {
            'name': 'mask',
            'description': 'Boundary mask generated from exterior vertices',
            'type': 'mask',
            'property': 'mask'
        },
        {
            'name': 'nuc_thresh',
            'description': 'Threshold applied to HSV-transformed image (S-channel) to segment the Hematoxylin channel.',
            'type': 'numeric',
            'min': 0,
            'max': 255
        },
        {
            'name': 'eosin_thresh',
            'description':'Threshold applied to HSV-transformed image (S-channel) to segment the Eosin channel.',
            'type': 'numeric',
            'min':0,
            'max': 255
        }
    ],
    output_spec = [
        {
            'name': 'Sub-Compartment Mask',
            'type': 'image',
            'description': 'Derived sub-compartment segmentation, 1 = Nuclei, 2 = Eosinophilic, 3 = Luminal Space'
        }
    ],
)


# Defining functions for finding interstitial distance transform
def find_tissue(image):

    # Mean of all channels/frames to make grayscale mask
    gray_mask = np.squeeze(np.mean(image,axis=-1))

    threshold_val = 1.15*threshold_otsu(gray_mask)
    tissue_mask = gray_mask <= threshold_val

    tissue_mask = remove_small_holes(tissue_mask,area_threshold=150)
    tissue_mask = remove_small_objects(tissue_mask,min_size=100)

    return tissue_mask

def make_interstitium_dt_map(annotations, image):

    non_spot_annotations = [i for i in annotations if not i['properties']['name']=='Spots']

    im_width = int(image.shape[1])
    im_height = int(image.shape[0])

    im_mask = np.zeros((im_height,im_width),dtype=np.uint8)
    for a in non_spot_annotations:
        for b in a['features']:
            a_mask = rasterio.features.rasterize([shape(b['geometry'])],out_shape = (im_height,im_width))
            im_mask += a_mask

    bin_d_mask = im_mask>0
    slide_tissue_mask = find_tissue(image)>0
    non_structure_tissue = 255*np.uint8(np.bitwise_and(slide_tissue_mask,~bin_d_mask))

    heat_data = ndi.distance_transform_edt(non_structure_tissue)

    # Colormapping distance transform and converting to RGBA
    c_map = cm.get_cmap('jet')

    heat_data -= np.min(heat_data)
    heat_data /= np.max(heat_data)

    rgba_heat = 255*np.uint8(c_map(np.uint8(255*heat_data)))
    a_heat = int(255*0.5)*np.ones((rgba_heat.shape[0],rgba_heat.shape[1],1),dtype=np.uint8)
    
    # Making all non-tissue regions transparent
    a_heat[heat_data==0] = 0
    rgba_heat[:,:,-1] = a_heat[:,:,0]

    rgba_heat = np.uint8(255*resize(rgba_heat,output_shape=(image.shape[0],image.shape[1],4)))

    image = Image.fromarray(image).convert('RGB')
    #image.putalpha(255)

    return image, Image.fromarray(rgba_heat).convert('RGBA')

from dash_extensions.enrich import html, Input, Output, State
from dash import dcc, ALL, ctx, exceptions, Patch
def make_interstitium_dt_output(output,output_index):

    tissue_img, dt_mask = output
    tissue_img.putalpha(255)

    main_figure = go.Figure(
        layout = {'margin': {'t':0,'b':0,'l':0,'r':0}}
    )
    main_figure.add_trace(go.Image(z=tissue_img))
    main_figure.add_trace(go.Image(z=dt_mask,opacity=0.5,colormodel='rgba'))

    tissue_img_base_div = html.Div([
        dcc.Graph(
            figure = main_figure,
            id = {'type': 'dt-overlay-fig','index': 0}
        )
    ])

    dt_component = html.Div([
        html.H5('Interstitial Distance Transform'),
        tissue_img_base_div,
        dcc.Slider(
            min = 0,
            max = 1,
            step = 0.1,
            value = 0.5,
            id = {'type': 'dt-overlay-slider','index': output_index}
        )
    ])

    return dt_component

def update_dt_opacity(slider_val):

    if not any([i['value'] is not None for i in ctx.triggered]):
        raise exceptions.PreventUpdate
    
    updated_transparency = Patch()
    updated_transparency['data'][1]['opacity'] = slider_val[0]

    return [updated_transparency]

interstitium_thickness = FUSIONFunction(
    title = 'Interstitial Distance',
    description = 'Display distance between annotated structures as a heatmap',
    function = lambda annotations, image: make_interstitium_dt_map(annotations, image),
    function_type = 'ROI',
    input_spec = [
        {
            'name': 'annotations',
            'description': 'Annotations between which to find thickness',
            'type': 'annotation'
        },
        {
            'name': 'image',
            'description': 'Tissue containing interstitium and annotated structures.',
            'type': 'image'
        }
    ],
    output_spec = [
        {
            'type': 'function',
            'function': lambda output, output_index: make_interstitium_dt_output(output,output_index)
        }
    ],
    output_callbacks = [
        {
            'inputs': [
                Input({'type': 'dt-overlay-slider','index': ALL},'value')
            ],
            'outputs': [
                Output({'type': 'dt-overlay-fig','index':ALL},'figure')
            ],
            'function': lambda slider_val: update_dt_opacity(slider_val)
        }
    ]
)




def main():

    # Grabbing first item from demo DSA instance
    #base_url = 'https://demo.kitware.com/histomicstk/api/v1'
    #item_id = ['5bbdeed1e629140048d01bcb','58b480ba92ca9a000b08c89d']
    base_url = os.environ.get('DSA_URL')

    item_id = [
        '6495a4e03e6ae3107da10dc5',
        '6495a4df3e6ae3107da10dc2'
    ] 

    # Starting the DSAHandler to grab information:
    dsa_handler = DSAHandler(
        girderApiUrl = base_url
    )
    
    vis_session = Visualization(
        tileservers=[dsa_handler.get_tile_server(i) for i in item_id],
        components = {
            'Visualization': [
                [
                    SlideMap(),
                    [
                        CustomFunction(
                            title = 'Test Functions',
                            description='Testing out different input/output types',
                            custom_function=[
                                buffer_shapes_component,
                                stain_mask_component,
                                interstitium_thickness
                            ]
                        ),
                        OverlayOptions()
                    ]
                ],
            ],
            'Dataset Builder': [
                dsa_handler.create_dataset_builder()
            ]
        }
    )

    vis_session.start()


if __name__=='__main__':
    main()





from fusion_tools.components.maps import (
    MapComponent, SlideMap, MultiFrameSlideMap, HybridSlideMap,
    SlideImageOverlay, ChannelMixer, LargeSlideMap, LargeMultiFrameSlideMap
)
from fusion_tools.components.tools import (
    OverlayOptions, PropertyViewer, PropertyPlotter, 
    HRAViewer, DataExtractor, GlobalPropertyPlotter,
    CustomFunction, FUSIONFunction
)
from fusion_tools.components.segmentation import FeatureAnnotation, BulkLabels
from fusion_tools.components.slide_annotation import SlideAnnotation, SlideAnnotationSchema



"""
Components for annotation of entire slides by users

"""

import os
import sys
import json
import geojson
import geopandas as gpd
import numpy as np
import pandas as pd
import textwrap
import re
import uuid
import threading
import zipfile
from shutil import rmtree
from copy import deepcopy

from typing_extensions import Union
from shapely.geometry import box, shape
import plotly.express as px
import plotly.graph_objects as go
from umap import UMAP

from PIL import Image, ImageOps

from io import BytesIO
import requests

# Dash imports
import dash
dash._dash_renderer._set_react_version('18.2.0')
import dash_leaflet as dl
import dash_leaflet.express as dlx
from dash import dcc, callback, ctx, ALL, MATCH, exceptions, Patch, no_update, dash_table
from dash.dash_table.Format import Format, Scheme
import dash_bootstrap_components as dbc
import dash_mantine_components as dmc
import dash_treeview_antd as dta
from dash_extensions.enrich import DashBlueprint, html, Input, Output, State, PrefixIdTransform, MultiplexerTransform
from dash_extensions.javascript import Namespace, arrow_function

# fusion-tools imports
from fusion_tools.visualization.vis_utils import get_pattern_matching_value
from fusion_tools.utils.shapes import (
    find_intersecting, 
    extract_geojson_properties, 
    process_filters_queries,
    detect_histomics,
    histomics_to_geojson,
    export_annotations
)
from fusion_tools.utils.images import get_feature_image, write_ome_tiff, format_intersecting_masks
from fusion_tools.utils.stats import get_label_statistics, run_wilcox_rank_sum
from fusion_tools import Tool, MultiTool

from fusion_tools.handler.dsa_handler import DSAHandler


class SlideAnnotationSchema:
    """Specification for SlideAnnotation schema

    .. code-block:: json

    {
        "dsa_url": "optional",
        "name": "required",
        "description": "optional",
        "users": ["optional"],
        "admins": ["optional"],
        "annotations": [
            {
                "name": "Example Text Label",
                "description": "",
                "type": "text",
                "roi": True,
                "editable": True
            },
            {
                "name": "Example Numeric Label",
                "description": "",
                "type": "numeric",
                "min": 0,
                "max": 100,
                "roi": False,
                "editable": True
            },
            {
                "name": "Example Options Label",
                "description": "",
                "type": "options",
                "options": [
                    "option 1",
                    "option 2",
                    "option 3"
                ],
                "multi": False,
                "roi": False,
                "editable": True
            }
        ]
    }

    """
    def __init__(self,
                 schema_data: dict):
        
        self.schema_data = schema_data

        # Checking dictionary for acceptable keys and validity
        allowed_keys = ["dsa_url", "name","description","users","admins","annotations","slides"]
        allowed_annotation_keys = ["name","description","type","options","multi","roi","editable","min","max"]

        allowed_key_vals = {
            "type": ["text","numeric","options","boolean","roi"],
            "editable": [True, False, "admins"],
            "multi": [True, False],
        }

    def to_dict(self):
        return self.schema_data


class SlideAnnotation(MultiTool):
    """Component used for assigning labels to slides, allows for importing other schema

    :param MultiTool: General class for tool which works on multiple slides at once
    :type MultiTool: None
    """
    def __init__(self,
                 handler: Union[None,DSAHandler] = None,
                 preload_schema: Union[None, str, dict, list, SlideAnnotationSchema] = None
                ):

        super().__init__()

        self.handler = handler
        self.preload_schema = preload_schema

        self.schemas = []

        if not self.preload_schema is None:
            local_schemas = self.load_local_schema(self.preload_schema)
        else:
            local_schemas = []
        
        if not self.handler is None:
            cloud_schemas = self.load_cloud_schema()
        else:
            cloud_schemas = []
        
        self.schemas = local_schemas+cloud_schemas

    def __str__(self):
        return 'Slide Annotation'
    
    def load(self, component_prefix: int):

        self.component_prefix = component_prefix
        self.title = 'Slide Annotation'
        self.blueprint = DashBlueprint(
            transforms = [
                PrefixIdTransform(prefix = f'{self.component_prefix}'),
                MultiplexerTransform()
            ]
        )

        self.get_callbacks()

    def get_callbacks(self):
        
        #TODO: Remaining callbacks to implement:
        # Callback for creating/updating annotation schema

        # Optional: Callback for inviting user to existing annotation schema
        # Optional: Callback for admin panel indicating other user's progress

        self.blueprint.callback(
            [
                Input({'type': 'slide-select-drop','index': ALL},'value')
            ],
            [
                State({'type': 'slide-annotation-schema-drop','index': ALL},'value'),
                State({'type': 'slide-annotation-input-info','index': ALL},'data'),
                State('anchor-vis-store','data')
            ],
            [
                Output({'type': 'slide-annotation-current-slide-div','index': ALL},'children'),
                Output({'type': 'slide-annotation-input','index': ALL},'value'),
                Output({'type': 'slide-annotation-input-info','index': ALL},'data'),
                Output({'type': 'slide-annotation-roi-input','index': ALL},'color')
            ]
        )(self.update_slide)

        self.blueprint.callback(
            [
                Input({'type':'slide-annotation-schema-drop','index': ALL},'value')
            ],
            [
                State({'type': 'slide-select-drop','index': ALL},'value'),
                State('anchor-vis-store','data')
            ],
            [
                Output({'type': 'slide-annotation-schema-parent-div','index':ALL},'children'),
                Output({'type': 'slide-annotation-download-button','index': ALL},'disabled'),
            ]
        )(self.update_schema)

        self.blueprint.callback(
            [
                Input({'type': 'slide-annotation-schema-refresh-icon','index': ALL},'n_clicks')
            ],
            [
                State('anchor-vis-store','data')
            ],
            [
                Output({'type': 'slide-annotation-schema-drop','index': ALL},'options')
            ]
        )(self.refresh_schemas)

        self.blueprint.callback(
            [
                Input({'type': 'slide-annotation-edit-input','index': ALL},'n_clicks')
            ],
            [
                State({'type': 'slide-annotation-input-info','index': ALL},'data')
            ],
            [
                Output({'type': 'slide-annotation-modal','index': ALL},'is_open'),
                Output({'type': 'slide-annotation-modal','index': ALL},'children')
            ]
        )(self.open_edit_modal)

        self.blueprint.callback(
            [
                Input({'type': 'slide-annotation-roi-input','index': ALL},'n_clicks')
            ],
            [
                State({'type': 'slide-annotation-input-info','index': ALL},'data'),
                State({'type': 'map-tile-layer','index': ALL},'url'),
                State({'type': 'map-tile-layer','index': ALL},'tileSize')
            ],
            [
                Output({'type': 'slide-annotation-modal','index': ALL},'is_open'),
                Output({'type': 'slide-annotation-modal','index': ALL},'children')
            ]
        )(self.open_roi_modal)

        self.blueprint.callback(
            [
                Input({'type': 'slide-annotation-submit-labels','index': ALL},'n_clicks')
            ],
            [
                State({'type': 'slide-annotation-input','index': ALL},'value'),
                State({'type': 'slide-annotation-input-info','index': ALL},'data'),
                State({'type': 'slide-annotation-schema-drop','index':ALL},'value'),
                State({'type': 'map-slide-information','index': ALL},'data'),
                State('anchor-vis-store','data')
            ],
            [
                Output({'type': 'slide-annotation-output-parent','index':ALL},'children'),
                Output({'type': 'slide-annotation-download-button','index': ALL},'disabled'),
                Output('anchor-vis-store','data')
            ]
        )(self.submit_labels)

        self.blueprint.callback(
            [
                Input({'type': 'slide-annotation-roi-done-button','index': ALL},'n_clicks')
            ],
            [
                State({'type': 'slide-annotation-edit-control','index': ALL},'geojson'),
                State({'type': 'slide-annotation-input-info','index': ALL},'data'),
                State({'type': 'map-slide-information','index': ALL},'data')
            ],
            [
                Output({'type': 'slide-annotation-input-info','index': ALL},'data'),
                Output({'type': 'slide-annotation-modal','index': ALL},'is_open'),
                Output({'type': 'slide-annotation-roi-input','index': ALL},'color')
            ]
        )(self.submit_roi)

        self.blueprint.callback(
            [
                Input({'type': 'slide-annotation-download-button','index': ALL},'n_clicks')
            ],
            [
                State({'type':'slide-annotation-schema-drop','index': ALL},'value'),
                State('anchor-vis-store','data')
            ],
            [
                Output({'type': 'slide-annotation-download-data','index': ALL},'data')
            ]
        )(self.download_annotations)

    def update_layout(self, session_data:dict, use_prefix:bool):

        # Components needed:
        # - Available Schema dropdown (shows existing schemas and New Schema)
        #   - Add refresh button for checking for cloud schema
        # - Current Slide info (shows name of current slide, id, any currently applied labels to that slide)
        #   - If this slide isn't added yet, have a button to "Add to Annotation Session" which reveals the annotation components
        #   - Underneath will be all schema components (rows of labeling components and "Submit" and "Clear" buttons at the bottom)
        # - Other slides labeled
        #   - If any others are added
        # - Download session data
        #   - Two files, one that is session metadata (schema) and the other which includes applied slide labels

        # Getting slide-annotation data from session 
        slide_annotation_data = session_data.get('data',{}).get('slide-annotation')
        if not slide_annotation_data is None:
            # Loading schemas from session data
            pass

        layout = html.Div([
            dbc.Card([
                dbc.CardBody([
                    dbc.Row(
                        dbc.Col(
                            html.H3('Slide Annotation')
                        )
                    ),
                    html.Hr(),
                    dbc.Row(
                        dbc.Col(
                            'Used for annotating whole slides following pre-specified schema'
                        )
                    ),
                    html.Hr(),
                    dbc.Modal(
                        id = {'type': 'slide-annotation-modal','index':0},
                        children = [],
                        is_open = False,
                        size = 'xl'
                    ),
                    dbc.Row([
                        dbc.Col(
                            dbc.Label(
                                'Select Annotation Schema: '
                            ),
                            md = 4
                        ),
                        dbc.Col(
                            dcc.Dropdown(
                                options = [
                                    {
                                        'label': n.schema_data['name'],
                                        'value': n.schema_data['name']
                                    }
                                    for n in self.schemas
                                ],
                                value = [],
                                multi = False,
                                id = {'type': 'slide-annotation-schema-drop','index': 0}
                            ),
                            md = 6
                        ),
                        dbc.Col([
                            html.A(
                                html.I(
                                    className = 'fa-solid fa-rotate fa-2x',
                                    n_clicks = 0,
                                    id = {'type': 'slide-annotation-schema-refresh-icon','index': 0}
                                )
                            ),
                            dbc.Tooltip(
                                target = {'type': 'slide-annotation-schema-refresh-icon','index': 0},
                                children = 'Click to refresh available schema'
                            )
                        ])
                    ]),
                    html.Hr(),
                    html.Div(
                        id = {'type': 'slide-annotation-schema-parent-div','index': 0},
                        children = [
                            'Select a schema to get started!'
                        ]
                    ),
                    html.Div(
                        children = [
                            dbc.Button(
                                'Download Annotations',
                                className = 'd-grid col-12 mx-auto',
                                color = 'success',
                                disabled = True,
                                n_clicks = 0,
                                id = {'type': 'slide-annotation-download-button','index': 0}
                            ),
                            dcc.Download(
                                id = {'type': 'slide-annotation-download-data','index': 0}
                            )
                        ],
                        style = {'marginTop': '5px'}
                    )
                ])
            ])
        ])

        if use_prefix:
            PrefixIdTransform(prefix = self.component_prefix).transform_layout(layout)

        return layout

    def gen_layout(self, session_data:dict):

        self.blueprint.layout = self.update_layout(session_data, use_prefix = False)

    def load_local_schema(self, schema):
        """Adding/preloading component with annotation schemas
        
        :param schema: Filepath, data, or object containing information describing the new schema
        :type schema: Union[str, list, dict, SlideAnnotationSchema]
        """
        
        if type(schema)==str:
            # Attempting to load schema from filepath
            with open(schema,'r') as f:
                schema = json.load(f)
                f.close()

        if type(schema)==dict:
            new_schema = [SlideAnnotationSchema(
                schema_data = schema
            )]
        
        elif type(schema)==list:
            new_schema = [
                SlideAnnotationSchema(
                    schema_data = s
                )
                for s in schema
            ]
        
        elif type(schema)==SlideAnnotationSchema:
            new_schema = [schema]

        else:
            new_schema = []
        
        return new_schema

    def load_cloud_schema(self):
        """Checking linked DSA instance for annotation session collection
        """
        cloud_schemas = []


        return cloud_schemas
    
    def refresh_schemas(self, clicked, session_data):

        if not any([i['value'] for i in ctx.triggered]):
            raise exceptions.PreventUpdate
        
        #TODO: check linked cloud instance for any new schemas added
        new_schema_options = [
            {
                'label': i.schema_data['name'],
                'value': i.schema_data['name']
            }
            for i in self.schemas
        ]

        new_schema_options += [
            {
                'label': 'New Schema',
                'value': 'New Schema'
            }
        ]

        return [new_schema_options]

    def update_slide(self, new_slide_index, schema_val, current_input_infos, session_data):
        

        if not any([i['value'] or i['value']==0 for i in ctx.triggered]):
            raise exceptions.PreventUpdate

        session_data = json.loads(session_data)
        new_slide = session_data['current'][get_pattern_matching_value(new_slide_index)]
        schema_val = get_pattern_matching_value(schema_val)

        new_slide_div = html.Div(
            children = [
                html.H5(f'Labeling for: {new_slide["name"]}')
            ]
        )

        # Checking if this slide has any labels        
        previous_labels = session_data.get('data',{}).get('slide-annotation',{}).get(schema_val,None)
        current_input_infos = [json.loads(i) for i in current_input_infos]

        if not previous_labels is None:
            slide_names = [i['Slide Name'] for i in previous_labels]
            if new_slide['name'] in slide_names:
                prev_input_vals = previous_labels[slide_names.index(new_slide['name'])]

                load_input_vals = [v for k,v in prev_input_vals.items() if not k in ['Slide Name','Slide ID'] and not '_ROI' in k]
                load_input_infos = [no_update if not i['roi'] else i | {'roi': prev_input_vals.get(f'{l}_ROI',True)} for i,l in zip(current_input_infos,[i for i in list(prev_input_vals.keys()) if not i in ['Slide Name','Slide ID']])]
                load_roi_input_colors = [no_update if not type(i['roi'])==dict else 'success' for i,l in zip(current_input_infos,[i for i in list(prev_input_vals.keys()) if not i in ['Slide Name','Slide ID']])]
            else:
                load_input_vals = [[] for i in range(len(ctx.outputs_list[1]))]
                load_input_infos = [no_update if not i['roi'] else i | {'roi': True} for i in current_input_infos]
                load_roi_input_colors = ['primary' if i['roi'] else 'secondary' for i in current_input_infos]
        else:
            load_input_vals = [[] for i in range(len(ctx.outputs_list[1]))]
            load_input_infos = [no_update if not i['roi'] else i | {'roi': True} for i in current_input_infos]
            load_roi_input_colors = ['primary' if i['roi'] else 'secondary' for i in current_input_infos]

        return [new_slide_div], load_input_vals, [json.dumps(i) if type(i)==dict else i for i in load_input_infos], load_roi_input_colors
    
    def update_schema(self, new_schema_val, current_slide_index, session_data):
        
        if not any([i['value'] for i in ctx.triggered]):
            raise exceptions.PreventUpdate
        
        session_data = json.loads(session_data)
        if not get_pattern_matching_value(current_slide_index) is None:
            current_slide = session_data['current'][get_pattern_matching_value(current_slide_index)]['name']
        else:
            current_slide = None

        new_schema_val = get_pattern_matching_value(new_schema_val)
        new_ann_components = self.make_annotation_components(new_schema_val, current_slide, session_data)

        # Checking if there is any data in the session for this schema
        schema_data = session_data.get('data',{}).get('slide-annotation',{}).get(new_schema_val)
        if not schema_data is None:
            disable_download_button = False
        else:
            disable_download_button = True

        return [new_ann_components], [disable_download_button]

    def make_annotation_components(self, schema_key, slide_name, session_data):

        if not schema_key in [i.schema_data['name'] for i in self.schemas]:
            return f'Schema: {schema_key} Not Found!'
        
        schema_index = [i.schema_data['name'] for i in self.schemas].index(schema_key)
        schema_info = self.schemas[schema_index].schema_data

        # Check if this schema has any data already in the session
        if 'slide-annotation' in session_data['data']:
            if schema_key in session_data['data']['slide-annotation']:
                schema_table = self.make_schema_label_table(session_data,schema_key)
                schema_data = session_data['data']['slide-annotation'][schema_key]
            else:
                schema_table = 'No labels added yet!'
                schema_data = None
        else:
            schema_table = 'No labels added yet!'
            schema_data = None

        if not schema_data is None:
            slide_names = [i['Slide Name'] for i in schema_data]
            if not slide_name is None:
                if slide_name in slide_names:
                    # First two values are Slide Name and Slide ID
                    slide_input_vals = schema_data[slide_names.index(slide_name)]
                else:
                    slide_input_vals = None
            else:
                slide_input_vals = None
        else:
            slide_input_vals = None


        schema_div = html.Div([
            dbc.Card([
                dbc.CardHeader(html.H4(schema_key)),
                dbc.CardBody([
                    dbc.Row(
                        schema_info.get('description')
                    ),
                    html.Hr(),
                    html.Div(
                        id = {'type': 'slide-annotation-current-slide-div','index': 0},
                        children = [
                            html.H5(f'Labeling for: {slide_name}') if not slide_name is None else ''
                        ]
                    ),
                    html.Div(
                        children = [
                            self.make_input_component(i,idx, slide_input_vals)
                            for idx,i in enumerate(schema_info.get('annotations',[]))
                        ]
                    ),
                    dbc.Row([
                        dbc.Button(
                            'Submit Labels',
                            className = 'd-grid col-12 mx-auto',
                            id = {'type': 'slide-annotation-submit-labels','index': 0},
                            n_clicks = 0,
                            color = 'primary',
                            disabled = False
                        )
                    ]),
                    html.Div(
                        id = {'type': 'slide-annotation-output-parent','index': 0},
                        children = [schema_table],
                        style = {
                            'marginTop': '10px',
                            'marginBottom': '10px',
                            'maxHeight': '20vh',
                            'overflow': 'scroll'
                        }
                    )
                ])
            ])
        ])

        PrefixIdTransform(prefix=f'{self.component_prefix}').transform_layout(schema_div)

        return [schema_div]

    def make_input_component(self, input_spec, input_index, slide_input_vals):

        
        roi_input_color = 'secondary' if not input_spec.get('roi',False) else 'primary'
        use_val = []
        if not slide_input_vals is None:
            if input_spec.get('roi',False):
                if f"{input_spec['name']}_ROI" in slide_input_vals:
                    slide_input_roi = slide_input_vals.get(f"{input_spec['name']}_ROI")
                    if type(slide_input_roi)==dict:
                        input_spec['roi'] = slide_input_roi
                        roi_input_color = 'success'
                    else:
                        roi_input_color = 'primary'
                else:
                    roi_input_color = 'primary'
            
            if slide_input_vals.get(input_spec['name'],False):
                use_val = slide_input_vals.get(input_spec['name'])
            else:
                use_val = []

        input_desc_column = [
            dbc.Row(html.H6(input_spec['name'])),
            dbc.Row(html.P(input_spec['description'])),
            dcc.Store(
                id = {'type': 'slide-annotation-input-info','index': input_index},
                data = json.dumps(input_spec),
                storage_type = 'memory'
            )
        ]

        edit_button = dbc.Button(
            children = [
                html.A(
                    html.I(
                        className = 'fa-solid fa-pen-to-square'
                    ),
                ),
                dbc.Tooltip(
                    target = {'type': 'slide-annotation-edit-input','index': input_index},
                    children = 'Edit Input Properties'
                )
            ],
            id = {'type': 'slide-annotation-edit-input','index': input_index},
            color = 'primary' if input_spec.get('editable',False) else 'secondary',
            n_clicks = 0,
            disabled = not input_spec.get('editable',False) 
        ) 

        roi_button = dbc.Button(
            children = [
                html.A(
                    html.I(
                        className = 'fa-solid fa-draw-polygon'
                    ),
                ),
                dbc.Tooltip(
                    target = {'type': 'slide-annotation-roi-input','index': input_index},
                    children = 'Draw ROI'
                )
            ],
            id = {'type': 'slide-annotation-roi-input','index': input_index},
            color = roi_input_color,
            n_clicks = 0,
            disabled= roi_input_color=='secondary'
        ) 

        if input_spec['type']=='text':
            input_component = html.Div([
                dbc.Row([
                    dbc.Col(input_desc_column,md=5),
                    dbc.Col([
                        dbc.InputGroup([
                            dbc.Input(
                                type = 'text',
                                value = use_val,
                                id = {'type': 'slide-annotation-input','index': input_index},
                            ),
                            roi_button,
                            edit_button                           
                        ])
                    ],md=7)
                ]),
                html.Hr()
            ])
        elif input_spec['type']=='boolean':

            level_names = input_spec.get('levels',['True','False'])
            if len(level_names)<2:
                level_names = ['True','False']

            input_component = html.Div([
                dbc.Row([
                    dbc.Col(input_desc_column,md=5),
                    dbc.Col([
                        dcc.RadioItems(
                            options = [
                                {'label': level_names[0], 'value': 1},
                                {'label': level_names[1], 'value': 0}
                            ],
                            inline = True,
                            value = use_val,
                            id = {'type': 'slide-annotation-input','index': input_index}
                        )
                    ],md=7)
                ])
            ])

        elif input_spec['type']=='numeric':
            input_component = html.Div([
                dbc.Row([
                    dbc.Col(input_desc_column,md=5),
                    dbc.Col([
                        dbc.InputGroup([
                            dbc.Input(
                                type = 'number',
                                value = use_val,
                                id = {'type': 'slide-annotation-input','index': input_index},
                            ),
                            roi_button,
                            edit_button                            
                        ])
                    ],md=7)
                ]),
                html.Hr()
            ])

        elif input_spec['type']=='options':
            #TODO: Find some workaround for the "multi" selection, all the className CSS options with dcc.Dropdown() didn't work here
            input_component = html.Div([
                dbc.Row([
                    dbc.Col(input_desc_column,md=5),
                    dbc.Col([
                        dbc.InputGroup([
                            dbc.Select(
                                options = input_spec['options'],
                                value = use_val,
                                id = {'type': 'slide-annotation-input','index': input_index}
                            ),
                            roi_button,
                            edit_button
                        ])
                    ],md=7)
                ]),
                html.Hr()
            ])

        return input_component

    def open_edit_modal(self, clicked, input_info):

        if not any([i['value'] for i in ctx.triggered]):
            raise exceptions.PreventUpdate
        
        input_info = json.loads(get_pattern_matching_value(input_info))
        
        return [True], [json.dumps(input_info,indent=4)]
    
    def open_roi_modal(self, clicked, input_info, tile_url, tile_size):

        if not any([i['value'] for i in ctx.triggered]):
            raise exceptions.PreventUpdate
        
        tile_url = get_pattern_matching_value(tile_url)
        tile_size = get_pattern_matching_value(tile_size)

        if any([i is None for i in [tile_url, tile_size]]):
            raise exceptions.PreventUpdate
        
        modal_children = [
            html.Div([
                dl.Map(
                    crs = 'Simple',
                    center = [-120,120],
                    zoom = 0,
                    children = [
                        dl.TileLayer(
                            url = tile_url,
                            tileSize=tile_size
                        ),
                        dl.FeatureGroup(
                            children = [
                                dl.EditControl(
                                    id = {'type': f'{self.component_prefix}-slide-annotation-edit-control','index': 0}
                                )
                            ]
                        )
                    ],
                    style = {'height': '40vh','width': '80%','margin': 'auto','display': 'inline-block'}
                ),
                dbc.Button(
                    'Done!',
                    className = 'd-grid col-12 mx-auto',
                    color = 'success',
                    n_clicks = 0,
                    id = {'type': f'{self.component_prefix}-slide-annotation-roi-done-button','index': 0}
                )
            ], style = {'padding': '10px 10px 10px 10px'})
        ]

        return [True], modal_children

    def edit_schema(self):
        pass

    def submit_labels(self, submit_clicked, input_vals, input_infos, schema_name, slide_information, session_data):
        
        if not any([i['value'] for i in ctx.triggered]):
            raise exceptions.PreventUpdate

        session_data = json.loads(session_data)
        schema_name = get_pattern_matching_value(schema_name)
        slide_information = json.loads(get_pattern_matching_value(slide_information))

        # If a slide hasn't been loaded yet
        if slide_information is None:
            raise exceptions.PreventUpdate
        elif len(list(slide_information.keys()))==0:
            raise exceptions.PreventUpdate

        input_infos = [json.loads(i) for i in input_infos]

        slide_annotation_data = session_data.get('data',{}).get('slide-annotation',{}).get(schema_name)

        slide_id = slide_information['metadata_url'].split('/')[-1]
        if slide_id == 'metadata':
            slide_id = slide_information['metadata_url'].split('/')[-2]

        this_slide_dict = {
            'Slide Name': slide_information['name'],
            'Slide ID': slide_id
        }

        this_slide_dict = this_slide_dict | {info['name']: val for info,val in zip(input_infos,input_vals)}
        if any([i['roi'] for i in input_infos]):
            for i in input_infos:
                if i['roi']:
                    this_slide_dict = this_slide_dict | {f'{i["name"]}_ROI': json.dumps(i['roi'])}

        if not slide_annotation_data is None:
            # If there is already recorded data for that schema
            # Checking if this slide is already present 
            current_ids = [i['Slide ID'] for i in slide_annotation_data]
            if this_slide_dict['Slide ID'] in current_ids:
                slide_annotation_data[current_ids.index(this_slide_dict['Slide ID'])] = this_slide_dict
            else:
                slide_annotation_data.append(this_slide_dict)
        else:
            # Initializing annotation schema
            if 'slide-annotation' in session_data['data']:
                session_data['data']['slide-annotation'][schema_name] = [this_slide_dict]
            else:
                session_data['data']['slide-annotation'] = {
                    schema_name: [this_slide_dict]
                }

        ann_schema_table = self.make_schema_label_table(session_data,schema_name)
        disable_download_button = False

        return [ann_schema_table], [disable_download_button], json.dumps(session_data)
    
    def make_schema_label_table(self, session_data, schema_name):

        schema_data = session_data.get('data',{}).get('slide-annotation',{}).get(schema_name)
        if not schema_data is None:
            ann_schema_df = pd.DataFrame.from_records(schema_data)

            ann_schema_table = dash_table.DataTable(
                id = {'type':f'{self.component_prefix}-slide-annotation-schema-table','index': 0},
                columns = [{'name':i,'id':i,'deletable':False,'selectable':True} for i in ann_schema_df.columns],
                data = ann_schema_df.to_dict('records'),
                fixed_columns={ 'headers': True, 'data': 1 },
                style_table={'minWidth': '100%'},
                style_cell={
                    # all three widths are needed
                    'minWidth': '250px', 'width': '250px', 'maxWidth': '250px',
                    'overflow': 'hidden',
                    'textOverflow': 'ellipsis',
                },
                tooltip_data = [
                    {
                        column: {'value': str(value),'type':'markdown'}
                        for column,value in row.items()
                    } for row in ann_schema_df.to_dict('records')
                ],
                tooltip_duration = None
            )
        else:
            ann_schema_table = html.Div()

        return ann_schema_table

    def submit_roi(self, done_clicked, edit_geojson, input_info, slide_information):
        
        if not any([i['value'] for i in ctx.triggered]):
            raise exceptions.PreventUpdate
        
        edit_geojson = get_pattern_matching_value(edit_geojson)

        n_inputs = len(input_info)
        input_info = json.loads(input_info[ctx.triggered_id['index']])

        slide_information = json.loads(get_pattern_matching_value(slide_information))
        scaled_geojson = geojson.utils.map_geometries(lambda g: geojson.utils.map_tuples(lambda c: (c[0]/slide_information['x_scale'],c[1]/slide_information['y_scale']),g),edit_geojson)

        input_info['roi'] = scaled_geojson

        update_infos = [no_update if not idx==ctx.triggered_id['index'] else json.dumps(input_info) for idx in range(n_inputs)]
        modal_open = [False]
        roi_button_color = [no_update if not idx==ctx.triggered_id['index'] else 'success' for idx in range(n_inputs)]

        return update_infos, modal_open, roi_button_color

    def download_annotations(self, button_clicked, schema_name, session_data):
        
        if not any([i['value'] for i in ctx.triggered]):
            raise exceptions.PreventUpdate
        
        schema_name = get_pattern_matching_value(schema_name)
        session_data = json.loads(session_data)

        annotation_data = session_data.get('data',{}).get('slide-annotation',{}).get(schema_name)
        if not annotation_data is None:
            annotation_df = pd.DataFrame.from_records(annotation_data)
            
            # Transforming the schema name so that it's a valid filename
            schema_save_name = re.sub(r'[^\w_.)( -]', '', schema_name)
            return [dcc.send_data_frame(annotation_df.to_csv,f'{schema_save_name}.csv')]
        else:
            return [no_update]

